@@ -4,30 +4,48 @@ from __future__ import unicode_literals
 import errno
 import io
 import os
-import mock
 import shutil
 import tempfile
+import sys
+import time
 import unittest
+import warnings
 
-from fs import osfs
-from fs import fsencode, fsdecode
-from fs.path import relpath
+from fs import osfs, open_fs
+from fs.path import relpath, dirname
 from fs import errors
-
 from fs.test import FSTestCases
 
 from six import text_type
 
+try:
+    from unittest import mock
+except ImportError:
+    import mock
+
 
 class TestOSFS(FSTestCases, unittest.TestCase):
     """Test OSFS implementation."""
 
+    @classmethod
+    def setUpClass(cls):
+        warnings.simplefilter("error")
+
+    @classmethod
+    def tearDownClass(cls):
+        warnings.simplefilter(warnings.defaultaction)
+
     def make_fs(self):
         temp_dir = tempfile.mkdtemp("fstestosfs")
         return osfs.OSFS(temp_dir)
 
     def destroy_fs(self, fs):
         self.fs.close()
+        try:
+            shutil.rmtree(fs.getsyspath("/"))
+        except OSError:
+            # Already deleted
+            pass
 
     def _get_real_path(self, path):
         _path = os.path.join(self.fs.root_path, relpath(path))
@@ -67,13 +85,45 @@ class TestOSFS(FSTestCases, unittest.TestCase):
 
     def test_not_exists(self):
         with self.assertRaises(errors.CreateFailed):
-            fs = osfs.OSFS("/does/not/exists/")
+            osfs.OSFS("/does/not/exists/")
 
-    @unittest.skipIf(osfs.sendfile is None, "sendfile not supported")
+    def test_expand_vars(self):
+        self.fs.makedir("TYRIONLANISTER")
+        self.fs.makedir("$FOO")
+        path = self.fs.getsyspath("$FOO")
+        os.environ["FOO"] = "TYRIONLANISTER"
+        fs1 = osfs.OSFS(path)
+        fs2 = osfs.OSFS(path, expand_vars=False)
+        self.assertIn("TYRIONLANISTER", fs1.getsyspath("/"))
+        self.assertNotIn("TYRIONLANISTER", fs2.getsyspath("/"))
+
+    def test_copy_preserve_time(self):
+        self.fs.makedir("foo")
+        self.fs.makedir("bar")
+        self.fs.create("foo/file.txt")
+        raw_info = {"details": {"modified": time.time() - 10000}}
+        self.fs.setinfo("foo/file.txt", raw_info)
+
+        namespaces = ("details", "modified")
+        src_info = self.fs.getinfo("foo/file.txt", namespaces)
+
+        self.fs.copy("foo/file.txt", "bar/file.txt", preserve_time=True)
+        self.assertTrue(self.fs.exists("bar/file.txt"))
+
+        dst_info = self.fs.getinfo("bar/file.txt", namespaces)
+        delta = dst_info.modified - src_info.modified
+        self.assertAlmostEqual(delta.total_seconds(), 0, places=2)
+
+    @unittest.skipUnless(osfs.sendfile, "sendfile not supported")
+    @unittest.skipIf(
+        sys.version_info >= (3, 8),
+        "the copy function uses sendfile in Python 3.8+, "
+        "making the patched implementation irrelevant",
+    )
     def test_copy_sendfile(self):
         # try copying using sendfile
         with mock.patch.object(osfs, "sendfile") as sendfile:
-            sendfile.side_effect = OSError(errno.ENOTSUP, "sendfile not supported")
+            sendfile.side_effect = OSError(errno.ENOSYS, "sendfile not supported")
             self.test_copy()
         # check other errors are transmitted
         self.fs.touch("foo")
@@ -115,7 +165,7 @@ class TestOSFS(FSTestCases, unittest.TestCase):
         finally:
             shutil.rmtree(dir_path)
 
-    @unittest.skipIf(not hasattr(os, "symlink"), "No symlink support")
+    @unittest.skipUnless(hasattr(os, "symlink"), "No symlink support")
     def test_symlinks(self):
         with open(self._get_real_path("foo"), "wb") as f:
             f.write(b"foobar")
@@ -142,3 +192,40 @@ class TestOSFS(FSTestCases, unittest.TestCase):
             with self.assertRaises(errors.InvalidCharsInPath):
                 with self.fs.open("13 – Marked Register.pdf", "wb") as fh:
                     fh.write(b"foo")
+
+    def test_consume_geturl(self):
+        self.fs.create("foo")
+        try:
+            url = self.fs.geturl("foo", purpose="fs")
+        except errors.NoURL:
+            self.assertFalse(self.fs.hasurl("foo"))
+        else:
+            self.assertTrue(self.fs.hasurl("foo"))
+
+        # Should not throw an error
+        base_dir = dirname(url)
+        open_fs(base_dir)
+
+    def test_complex_geturl(self):
+        self.fs.makedirs("foo/bar ha")
+        test_fixtures = [
+            # test file, expected url path
+            ["foo", "foo"],
+            ["foo-bar", "foo-bar"],
+            ["foo_bar", "foo_bar"],
+            ["foo/bar ha/barz", "foo/bar%20ha/barz"],
+            ["example b.txt", "example%20b.txt"],
+            ["exampleㄓ.txt", "example%E3%84%93.txt"],
+        ]
+        file_uri_prefix = "osfs://"
+        for test_file, relative_url_path in test_fixtures:
+            self.fs.create(test_file)
+            expected = file_uri_prefix + self.fs.getsyspath(relative_url_path).replace(
+                "\\", "/"
+            )
+            actual = self.fs.geturl(test_file, purpose="fs")
+
+            self.assertEqual(actual, expected)
+
+    def test_geturl_return_no_url(self):
+        self.assertRaises(errors.NoURL, self.fs.geturl, "test/path", "upload")
